#ifndef PIXEL_T_H
#define PIXEL_T_H
#include <cstdint>
typedef uint8_t pixel8_t;
typedef uint8_t ppixel24_t[3];
typedef pixel8_t *ppixel8_t;
typedef uint16_t pixel16_t;
typedef pixel16_t *ppixel16_t;
typedef uint32_t pixel32_t;
typedef pixel32_t *ppixel32_t;
#endif
